/* @bopen.c - buffered file library
 */
#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <assert.h>
#include <unistd.h>
#include <errno.h>
#ifdef FOO
#include <macros.h>
#endif
#ifdef XENIX
#include "xenix.h"
#endif
#ifdef DOS
#include "dos.h"
#endif
#include "porttypes.h"
#include "gen.h"
/*
#ifndef PORT
#	include "file.h"
#	include "filesys.h"
#else
	typedef int Oflag;
#endif
*/
#include "bopen.h"

typedef struct
	{
	char *buf; /* buffer */
	int size;  /* bytes in buffer */
	int pos;   /* position in buffer */
	int mode;  /* last access mode */
	FD realfd;
	Bool buffered; /* use buffering now? */
	Functionmap *map;
	} Filebuf;
/* modes */
#define M_READ 1
#define M_WRITE 2
#define MAXBUF (sizeof(bufmain) / sizeof(Filebuf *))
#define BUFFERSIZE BUFSIZ


static Filebuf *bufmain[] = {
NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL
};
static Bool initialized = NO;

#define binit() \
	{ \
	if (!initialized) \
		{ \
		initialized = YES; \
		atexit(allflush); \
		} \
	}

/* b@bopen - open a file, possibly buffered
 */
BFD bopen(char *path, Oflag oflag, Mode pmode, Functionmap *map)
	{
	BFD rtn;
	FD fd;

#	ifdef DEB
	errfmt("bopen path '%s'\n", path);
#	endif
	if ((fd = (*map->open)(path, oflag & ~O_BUFFER, pmode)) < 0)
		rtn = -ERRNO;
	else if (0 <= (rtn = bset(fd, map)) && (oflag & O_BUFFER))
		buffer(rtn);
	return rtn;
	}

/* b@allflush - flush buffers associated with all files
 */
void allflush(void)
	{
	FD i;

#	ifdef DEB
	errfmt("allflush\n");
#	endif
	for (i = 0; i < MAXBUF; i++)
		if (bufmain[i])
			flush(i);
	}

BFD bufferset(FD fd, Functionmap *map)
	{
	BFD rtn;

#	ifdef DEB
	errfmt("bufferset %d\n", fd);
#	endif
	rtn = bset(fd, map);
	buffer(rtn);
	return rtn;
	}

/* b@bset - make the file associated with fd buffered
 */
BFD bset(FD fd, Functionmap *map)
	{
	BFD rtn;
	Filebuf *filebuf = NULL;

#	ifdef DEB
	errfmt("bset:\n");
#	endif
	binit();
	if ((rtn = getfd((void **)bufmain, MAXBUF)) < 0)
		rtn = -ERRNO;
	else if ((filebuf = (Filebuf *)malloc(sizeof(Filebuf))) == NULL)
		rtn = -ERRNO;
	else if ((filebuf->buf = (char *)malloc(BUFFERSIZE)) == NULL)
		{
		free(filebuf);
		rtn = -ERRNO;
		}
	else
		{
		bufmain[rtn] = filebuf;
		filebuf->size = 0;
		filebuf->pos = 0;
		filebuf->mode = M_READ;
		filebuf->map = map;
		filebuf->buffered = NO;
		filebuf->realfd = fd;
		}
#	ifdef DEB
	errfmt("bset: rtn %d\n", rtn);
#	endif
	return rtn;
	}
/* b@bfd_to_fd - make the file associated with fd buffered
 */
FD bfd_to_fd(BFD bfd)
	{
	FD rtn;
	Filebuf *filebuf;

	if ((filebuf = (Filebuf *)getobj((void **)bufmain, MAXBUF, bfd)) == NULL)
		rtn = -ERRNO;
	else
		{
		rtn = filebuf->realfd;
		}
#	ifdef DEB
	errfmt("bfd_to_fd: bfd %d fd %d\n", bfd, rtn);
#	endif
	return rtn;
	}

/* b@buffer - make the file associated with fd buffered
 */
int buffer(BFD bfd)
	{
	int rtn = 0;
	Filebuf *filebuf;

#	ifdef DEB
	errfmt("buffer: fd %d\n", bfd);
#	endif
	if ((filebuf = (Filebuf *)getobj((void **)bufmain, MAXBUF, bfd)) == NULL)
		rtn = -ERRNO;
	else
		{
		if (!filebuf->buffered)
			{
			filebuf->size = 0;
			filebuf->pos = 0;
			filebuf->mode = M_READ;
			filebuf->buffered = YES;
			}
		}
#	ifdef DEB
	errfmt("buffer:end fd %d\n", bfd);
#	endif
	return rtn;
	}

/* b@bcreat - creat a file, in buffered mode 
 */
BFD bcreat(char *path, Mode pmode, Functionmap *map)
	{
	return bopen(path, O_WRONLY | O_CREAT | O_TRUNC | O_BUFFER, pmode, map);
	}

/* b@bseek - seek on a buffered file
 */
Seek bseek(BFD bfd, Seek seek, int whence)
	{
	Seek rtn;
	int add = 0;
	Filebuf *filebuf;

#	ifdef DEB
	errfmt("bseek fd %d seek %ld whence %d\n", bfd, seek, whence);
#	endif
	if ((filebuf = (Filebuf *)getobj((void **)bufmain, MAXBUF, bfd)) == NULL)
		rtn = -ERRNO;
	else
		{
		flush(bfd);
		if (filebuf->mode == M_READ)
			add = filebuf->pos - filebuf->size; /* this looks wrong */
		switch (whence)
			{
		case SEEK_SET:
		case SEEK_END:
		    rtn = (*filebuf->map->seek)(filebuf->realfd, seek, whence);
		    break;
		case SEEK_CUR:
		    rtn = (*filebuf->map->seek)(filebuf->realfd,
										seek - (filebuf->size - filebuf->pos),
										whence);
			/* does this work for negatives? */
				break;
		default:
	    	rtn = -(errno = EINVAL);
			}
		if (0 < rtn)
			{
			filebuf->pos = 0;
			filebuf->size = 0;
			}
		}
	return rtn;
	}

/* b@bunset - close a possibly buffered file
 */
int bunset(BFD bfd)
	{
	int rtn = 0;

#	ifdef DEB
	errfmt("bunset\n");
#	endif
	if (0 <= (rtn = unbuffer(bfd)))
		{
		Filebuf *filebuf;
		if ((filebuf = (Filebuf *)getobj((void **)bufmain, MAXBUF, bfd)) == NULL)
			rtn = -ERRNO;
		else
			{
			free(filebuf->buf);
			free(filebuf);
			bufmain[bfd] = NULL;
			}
		}
	return rtn;
	}

/* b@bclose - close a possibly buffered file
 */
int bclose(BFD bfd)
	{
	int rtn = 0;

#	ifdef DEB
	errfmt("bclose\n");
#	endif
	if (0 <= (rtn = unbuffer(bfd)))
		{
		Filebuf *filebuf;
		if ((filebuf = (Filebuf *)getobj((void **)bufmain, MAXBUF, bfd)) == NULL)
			rtn = -ERRNO;
		else
			{
			rtn = (*filebuf->map->close)(filebuf->realfd);
			bunset(bfd);
			}
		}
	return rtn;
	}

/* b@unbuffer - remove buffering from a file
 */
int unbuffer(BFD bfd)
	{
	int rtn = 0;
	Filebuf *filebuf;

#	ifdef DEB
	errfmt("unbuffer\n");
#	endif
	if ((filebuf = (Filebuf *)getobj((void **)bufmain, MAXBUF, bfd)) == NULL)
		rtn = -ERRNO;
	else
		{
		flush(bfd);
		filebuf->buffered = NO;
		}
	return rtn;
	}

/* b@flush - flush buffers for file associated with fd
 */
int flush(BFD bfd)
	{
	int rtn = 0;
	Filebuf *filebuf;

#	ifdef DEB
	errfmt("flush: %d fd\n", bfd);
#	endif
	if ((filebuf = (Filebuf *)getobj((void **)bufmain, MAXBUF, bfd)) == NULL)
		rtn = -ERRNO;
	else
		{
		if (filebuf->buffered && filebuf->mode == M_WRITE)
			{
			(*filebuf->map->write)(filebuf->realfd, filebuf->buf, filebuf->pos);
			filebuf->pos = 0;
			filebuf->mode = M_READ;
			}
		}
	return rtn;
	}

/* b@bread - read from a buffered file
 */
int bread(BFD bfd, void *buf, unsigned int nbyte)
	{
	int rtn = 0;
	int total = 0;
	int tran;
	Filebuf *filebuf;

#	ifdef DEB
	errfmt("bread: fd %d nbyte %d\n", bfd, nbyte);
#	endif
	if ((filebuf = (Filebuf *)getobj((void **)bufmain, MAXBUF, bfd)) == NULL)
		rtn = -ERRNO;
	else
		{
		if (filebuf->buffered)
			{
			if (filebuf->mode == M_WRITE)
				flush(bfd);
			while (total < nbyte)
				{
#			ifdef DEB
				errfmt("bread:1 pos %d size %d\n", filebuf->pos, filebuf->size);
#			endif
				if (filebuf->pos == filebuf->size)
					{
#			ifdef DEB
					errfmt("bread:2 pos %d size %d\n", filebuf->pos, filebuf->size);
#			endif
					flush(bfd);
					filebuf->pos = 0;
					if ((filebuf->size = (*filebuf->map->read)(filebuf->realfd, filebuf->buf, BUFFERSIZE)) <= 0)
						break;
					}
#			ifdef DEB
				errfmt("bread:3 pos %d size %d BUF %d nbyte %d total %d\n", filebuf->pos, filebuf->size, BUFFERSIZE, nbyte, total);
#			endif
				tran = min(nbyte - total, filebuf->size - filebuf->pos);
				memcpy((char *)buf + total, filebuf->buf + filebuf->pos, tran);
				total += tran;
				filebuf->pos += tran;
#			ifdef DEB
				errfmt("bread:4 pos %d size %d tran %d total %d\n", filebuf->pos, filebuf->size, tran, total);
#			endif
				}
			}
		else
			total = (*filebuf->map->read)(filebuf->realfd, buf, nbyte);
		rtn = total;
		}
#	ifdef DEB
	errfmt("bread: rtn %d\n", total);
#	endif
	return rtn;
	}

/* b@bwrite - write to a buffered file
 */
int bwrite(BFD bfd, void *buf, unsigned int nbyte)
	{
	int total = 0;
	int tran;
	Filebuf *filebuf;

#	ifdef DEB
	errfmt("bwrite: fd %d nbyte %d\n", bfd, nbyte);
#	endif
	if ((filebuf = (Filebuf *)getobj((void **)bufmain, MAXBUF, bfd)) == NULL)
		return -ERRNO;
	if (filebuf->buffered)
		{
		while (total < nbyte)
			{
			filebuf->mode = M_WRITE;
#			ifdef DEB
			errfmt("bwrite: pos %d size %d\n", filebuf->pos, filebuf->size);
#			endif
			tran = min(nbyte - total, BUFFERSIZE - filebuf->pos);
			memcpy(filebuf->buf + filebuf->pos, (char *)buf + total, tran);
			total += tran;
			filebuf->pos += tran;
			if (filebuf->pos == BUFFERSIZE && 0 < filebuf->pos)
				flush(bfd);
			}
		}
	else
		total = (*filebuf->map->write)(filebuf->realfd, buf, nbyte);
	return total;
	}

