/* @onintr.c - capture interrupts
 */
#ifdef sparc
#define __EXTENSIONS__ 1
#endif
#ifdef linux
/* #define __need_sigset_t */
#endif
#include <unistd.h>
#include <signal.h>
#include <sys/signal.h>
#ifdef linux
/*#define _SIGNAL_H
#include <signum.h> */
#define sigset_t __sigset_t
#include <sigaction.h>
#endif
#include "gen.h"

#define MAXSIGSTACK 10
typedef void (fptr)(void);
fptr *thesigstack[MAXSIGSTACK]; 
int numsigstack = 0;
int defer_interrupt = 0;
Bool interrupted = NO;

int poss_signals[] = {SIGABRT, SIGALRM, SIGFPE, SIGHUP, SIGILL, SIGINT, 
	SIGPIPE, SIGQUIT, SIGSEGV, SIGTERM, SIGUSR1, SIGUSR2};
#define num_poss_signals (sizeof(poss_signals)/sizeof(int))
int my_signals[] = {SIGABRT, SIGFPE, SIGHUP, SIGILL, SIGINT, 
	SIGPIPE, SIGQUIT, SIGSEGV, SIGTERM};
#define num_my_signals (sizeof(poss_signals)/sizeof(int))

Bool stored_old = NO;
struct sigaction old_set[num_poss_signals];
sigset_t old_procmask;

void set_signals(struct sigaction * sets)
{
	int i;
	struct sigaction act;
	act.sa_handler = interrupt;
	sigemptyset(&act.sa_mask);
	act.sa_flags = 0;
	for (i = 0; i < num_poss_signals; i++)
		sigaction(poss_signals[i], &sets[i], NULL);
}

void interrupt_exit(void)
	{
#	ifdef DEB
	errfmt("INTERRUPT-_exit:beg\n");
#	endif
	while(numsigstack--)
		thesigstack[numsigstack]();
#ifdef NEVER		
		(*((void (*)())(thesigstack[numsigstack])))();
#endif	
#	ifdef DEB
	errfmt("INTERRUPT-_exit:end\n");
#	endif
	/*
	_exit(E_SIGNAL);
	*/
	set_signals(&old_set[0]);
	kill(getpid(), SIGTERM);
	}

void interrupt(void)
	{
#	ifdef DEB
	errfmt("INTERRUPT!!!\n");
#	endif
/*
	if (defer_interrupt == 0)
		interrupt_exit();
	interrupted = YES;
*/
	interrupt_exit();
	}


void push_interrupt(void)
	{
	if (defer_interrupt == 0)
		{
		int i;
		sigset_t set;
		sigfillset(&set);
		sigprocmask(SIG_BLOCK, &set, &old_procmask);
		}
	defer_interrupt++;
	}

void pop_interrupt(void)
	{
	defer_interrupt--;
	if (defer_interrupt == 0)
		sigprocmask(SIG_BLOCK, &old_procmask, NULL);
/*
	if (interrupted && defer_interrupt == 0)
		interrupt_exit();
*/
	}


void store_old_signals(struct sigaction * sets)
{
	int i;
	for (i = 0; i < num_poss_signals; i++)
		sigaction(poss_signals[i], NULL, &sets[i]);
}


void onintr(void (*pfn)(void))
	{
	register void (*pf)(void);
	struct sigaction act;

	if (pfn == NULL)
		{
		pf = SIG_IGN;
		numsigstack = 0;
		}
	else if (pfn == SIG_DFL)
		{
		pf = SIG_DFL;
		numsigstack = 0;
		}
	else
		{
		if (numsigstack == MAXSIGSTACK)
			{
			write(2, "onintr: SIGNAL STACK OVERFLOW!!!\n", 33);
			exit(E_INTERN);
			}
		else
			{
			thesigstack[numsigstack++] = pfn;
			pf = interrupt;
			}
		}
#ifdef OLD_SIGNALS
	if (signal(SIGHUP, SIG_IGN) != SIG_IGN)
		signal(SIGHUP, pf);
	if (signal(SIGINT, SIG_IGN) != SIG_IGN)
		signal(SIGINT, pf);
	if (signal(SIGPIPE, SIG_IGN) != SIG_IGN)
		signal(SIGPIPE, pf);
#else
	if (!stored_old)
		{
		struct sigaction act;
		int i;

		store_old_signals(&old_set[0]);
		stored_old = YES;

		act.sa_handler = pf;
		sigemptyset(&act.sa_mask);
		act.sa_flags = 0;
		for (i = 0; i < num_my_signals; i++)
			sigaction(my_signals[i], &act, NULL);

		}
#endif
	}
